import asyncio

from abc import ABC, abstractmethod
from asyncio import CancelledError, Lock
from collections.abc import AsyncIterator
from typing import Self, TypeVar


T = TypeVar("T")


class AioplusIterator(ABC, AsyncIterator[T]):
    """An asynchronous iterator."""

    __slots__: tuple[str, ...] = (
        "_aioplus_cancelled_flg",
        "_aioplus_exception_flg",
        "_aioplus_finished_flg",
        "_aioplus_lock",
    )

    def __aiter__(self) -> Self:
        """Return self."""
        return self

    async def __anext__(self) -> T:
        """Return the next item."""
        if not hasattr(self, "_aioplus_lock"):
            self._aioplus_lock = Lock()

        async with self._aioplus_lock:
            if getattr(self, "_aioplus_cancelled_flg", False):
                detail = "cannot iterate after cancellation"
                raise StopAsyncIteration(detail)

            if getattr(self, "_aioplus_exception_flg", False):
                detail = "cannot iterate after exception(-s)"
                raise StopAsyncIteration(detail)

            if getattr(self, "_aioplus_finished_flg", False):
                detail = "cannot iterate after exhausting"
                raise StopAsyncIteration(detail)

            try:
                # Move to the next task!
                await asyncio.sleep(0.0)

                item = await self.__aioplus__()

            except StopAsyncIteration:
                self._aioplus_finished_flg = True
                raise

            except CancelledError:
                self._aioplus_cancelled_flg = True
                raise

            except BaseException:
                self._aioplus_exception_flg = True
                raise

        return item

    @abstractmethod
    async def __aioplus__(self) -> T:
        """Return the next item."""



import time

from collections.abc import Callable, Iterable, Iterator
from concurrent.futures import Future, ThreadPoolExecutor
from threading import Lock
from types import TracebackType
from typing import Any, Literal, ParamSpec, Self, TypeVar


P = ParamSpec("P")
R = TypeVar("R")


class CallerThreadExecutor(ThreadPoolExecutor):
    """An executor that uses the caller thread.

    Examples
    --------
    >>> executor = CallerThreadExecutor()
    >>> loop = asyncio.new_event_loop()
    >>> loop.set_default_executor(executor)

    See Also
    --------
    :class:`concurrent.futures.ThreadPoolExecutor`
    """

    __slots__ = ("_futures_count", "_futures_lock", "_shutdown", "_shutdown_lock")

    def __init__(
        self,
        max_workers: int | None = None,
        thread_name_prefix: str = "",
        initializer: Callable[..., Any] | None = None,
        initargs: tuple[Any, ...] = (),
    ) -> None:
        """Initialize the object.

        Parameters
        ----------
        max_workers : int, optional
            This parameter does not affect the behavior of the executor.

        thread_name_prefix : str
            This parameter does not affect the behavior of the executor.

        initializer : Callable[..., Any], optional
            This parameter does not affect the behavior of the executor.

        initargs : tuple[Any, ...]
            This parameter does not affect the behavior of the executor.

        Notes
        -----
        * Parameters are never used and serve as placeholders to comply with the interface of
          `concurrent.futures.ThreadPoolExecutor` (Liskov Substitution Principle).
        """
        if max_workers is not None and not isinstance(max_workers, int):
            detail = "'max_workers' must be 'int' or 'None'"
            raise TypeError(detail)

        if max_workers is not None and max_workers <= 0:
            detail = "'max_workers' must be positive"
            raise ValueError(detail)

        if not isinstance(thread_name_prefix, str):
            detail = "'thread_name_prefix' must be 'str'"
            raise TypeError(detail)

        if initializer is not None and not callable(initializer):
            detail = "'initializer' must be 'Callable' or 'None'"
            raise TypeError(detail)

        if not isinstance(initargs, tuple):
            detail = "'initargs' must be 'tuple'"
            raise TypeError(detail)

        self._futures_count: int = 0
        self._futures_lock = Lock()
        self._shutdown: bool = False
        self._shutdown_lock = Lock()

    def submit(self, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs) -> Future[R]:
        """Schedules the callable, ``fn``, to be executed.

        Parameters
        ----------
        fn : Callable[P, R]
            Callable.

        *args : P.args
            Positional arguments.

        **kwargs : P.kwargs
            Keyword arguments.

        Returns
        -------
        Future[R]
            Future.

        Notes
        -----
        * Submitted tasks are executed immediately.
        """
        if not callable(fn):
            detail = "'fn' must be 'Callable'"
            raise TypeError(detail)

        with self._shutdown_lock:
            if self._shutdown:
                detail = "cannot schedule new futures after shutdown"
                raise RuntimeError(detail)

            with self._futures_lock:
                self._futures_count += 1

        future: Future[R] = Future()
        try:
            result = fn(*args, **kwargs)
        except BaseException as exception:
            future.set_exception(exception)
        else:
            future.set_result(result)

        with self._futures_lock:
            self._futures_count -= 1

        return future

    def map(
        self,
        fn: Callable[..., R],
        *iterables: Iterable[Any],
        timeout: float | None = None,
        chunksize: int = 1,
        buffersize: int | None = None,
    ) -> Iterator[R]:
        """Map the callable to the iterables.

        Parameters
        ----------
        fn : Callable[P, R]
            Callable.

        *iterables : Iterable[P]
            Iterables to map the callable to.

        timeout : float, optional
            Timeout.

        chunksize : int
            This parameter does not affect the behavior of the executor.

        buffersize : int, optional
            This parameter does not affect the behavior of the executor.

        Returns
        -------
        Iterator[R]
            Iterator.

        Notes
        -----
        * This method submits lazily to avoid blocking the caller thread on infinite iterators;
        * If ``timeout`` is specified, then :exc:`TimeoutError` is always raised. The executor uses
          the caller thread, so there is no way to enforce the timeout.
        """
        if not callable(fn):
            detail = "'fn' must be 'Callable'"
            raise TypeError(detail)

        for iterable in iterables:
            if not isinstance(iterable, Iterable):
                detail = "'*iterables' must be 'Iterable'"
                raise TypeError(detail)

        if timeout is not None and not isinstance(timeout, float):
            detail = "'timeout' must be 'float' or 'None'"
            raise TypeError(detail)

        if not isinstance(chunksize, int):
            detail = "'chunksize' must be 'int'"
            raise TypeError(detail)

        if chunksize <= 0:
            detail = "'chunksize' must be positive"
            raise ValueError(detail)

        if buffersize is not None and not isinstance(buffersize, int):
            detail = "'buffersize' must be 'int' or 'None'"
            raise TypeError(detail)

        if buffersize is not None and buffersize <= 0:
            detail = "'buffersize' must be positive"
            raise ValueError(detail)

        def iterator() -> Iterator[R]:
            """Iterate the results."""
            if timeout is not None:
                time.sleep(timeout)
                raise TimeoutError

            for args in zip(*iterables, strict=False):
                future = self.submit(fn, *args)
                yield future.result()

        return iterator()

    def shutdown(
        self,
        wait: bool = True,  # noqa: FBT001, FBT002
        *,
        cancel_futures: bool = False,
    ) -> None:
        """Signal the executor that it should free resources.

        Parameters
        ----------
        wait : bool
            Waiting for futures.

        cancel_futures : bool
            This parameter does not affect the behavior of the executor.

        Notes
        -----
        * If ``wait`` is :obj:`True` then this method will not return until all the pending futures
          are done executing and the resources associated with the executor have been freed. If
          ``wait`` is :obj:`False` then this method will return immediately and the resources
          associated with the executor will be freed when all pending futures are done executing.
        """
        if not isinstance(wait, bool):
            detail = "'wait' must be 'bool'"
            raise TypeError(detail)

        if not isinstance(cancel_futures, bool):
            detail = "'cancel_futures' must be 'bool'"
            raise TypeError(detail)

        with self._shutdown_lock:
            self._shutdown = True

        if wait:
            while self._futures_count > 0:
                pass  # Just a spin-wait

    def __enter__(self) -> Self:
        """Enter the context.

        Returns
        -------
        Self
            Self.
        """
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> Literal[False]:
        """Shutdown the executor and wait for all futures to complete.

        Parameters
        ----------
        exc_type : type[BaseException] | None
            The exception type.

        exc_val : BaseException | None
            The exception instance.

        exc_tb : TracebackType | None
            The traceback object.

        Returns
        -------
        Literal[False]
            Reraise exceptions.
        """
        self.shutdown(wait=True)
        return False



from collections.abc import AsyncIterator
from dataclasses import dataclass
from types import EllipsisType
from typing import overload

from aioplus.internal.utils.abc import AioplusIterator


@overload
def arange(stop: int, /) -> AsyncIterator[int]: ...


@overload
def arange(start: int, stop: int, /) -> AsyncIterator[int]: ...


@overload
def arange(start: int, stop: int, step: int, /) -> AsyncIterator[int]: ...


def arange(
    start: int,
    stop: int | EllipsisType = ...,
    step: int | EllipsisType = ...,
    /,
) -> AsyncIterator[int]:
    """Return a sequence of numbers.

    Parameters
    ----------
    start : int
        Start.

    stop : int, unset
        Stop.

    step : int, unset
        Step.

    Returns
    -------
    AsyncIterator[int]
        Iterator.

    Examples
    --------
    >>> [num async for num in arange(23)]
    [0, 1, 2, 3, 4, ..., 19, 20, 21, 22]

    See Also
    --------
    :func:`range`
    """
    if not isinstance(start, int):
        detail = "'start' must be 'int'"
        raise TypeError(detail)

    if (stop is not ...) and not isinstance(stop, int):
        detail = "'stop' must be 'int'"
        raise TypeError(detail)

    if (step is not ...) and not isinstance(step, int):
        detail = "'step' must be 'int'"
        raise TypeError(detail)

    if (stop is ...) and (step is not ...):
        detail = "'stop' must be 'int'"
        raise ValueError(detail)

    if stop is ...:
        stop = start
        start = 0
        step = 1

    if step is ...:
        step = 1

    if not step:
        detail = "'step' must not be zero"
        raise ValueError(detail)

    return ArangeIterator(start, stop, step)


@dataclass(repr=False)
class ArangeIterator(AioplusIterator[int]):
    """An asynchronous iterator."""

    start: int
    stop: int
    step: int

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._next = self.start

    async def __aioplus__(self) -> int:
        """Return the next item."""
        if self.step > 0 and self._next >= self.stop:
            raise StopAsyncIteration

        if self.step < 0 and self._next <= self.stop:
            raise StopAsyncIteration

        value = self._next
        self._next += self.step

        return value



from collections.abc import AsyncIterable

from aioplus.internal.utils.typing import SupportsBool


async def aany(aiterable: AsyncIterable[SupportsBool], /) -> bool:
    """Return :obj:`True` if any items of ``aiterable`` evaluate to :obj:`True`.

    Parameters
    ----------
    aiterable : AsyncIterable[SupportsBool]
        Iterable.

    Returns
    -------
    :class:`bool`
        :obj:`True` if any item evaluates to :obj:`True`, otherwise :obj:`False`.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await aany(aiterable)
    True

    Notes
    -----
    * If ``aiterable`` is empty, then :obj:`False` is returned.

    See Also
    --------
    :func:`any`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    async for item in aiterable:
        if item:
            return True

    return False



from collections import deque
from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


def acycle(aiterable: AsyncIterable[T], /) -> AsyncIterator[T]:
    """Repeat ``aiterable`` items indefinitely.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [num async for num in acycle(aiterable)]
    [0, 1, ..., 22, 23, 0, 1, ..., 22, 23, ...]

    See Also
    --------
    :func:`itertools.cycle`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    return AcycleIterator(aiterator)


@dataclass(repr=False)
class AcycleIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    aiterator: AsyncIterator[T]

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._deque: deque[T] = deque()
        self._initialized_flg = False

    async def __aioplus__(self) -> T:
        """Return the next item."""
        if not self._initialized_flg:
            try:
                item = await anext(self.aiterator)
            except StopAsyncIteration:
                self._initialized_flg = True
            else:
                self._deque.append(item)
                return item

        if not self._deque:
            raise StopAsyncIteration

        item = self._deque.popleft()
        self._deque.append(item)

        return item



from collections.abc import AsyncIterable
from typing import Any, TypeVar, overload


T = TypeVar("T")
D = TypeVar("D")


@overload
async def afirst(aiterable: AsyncIterable[T], /) -> T: ...


@overload
async def afirst(aiterable: AsyncIterable[T], /, *, default: D) -> T | D: ...


async def afirst(aiterable: AsyncIterable[Any], /, *, default: Any = ...) -> Any:
    """Return the first item of ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    default : D, unset
        Default.

    Returns
    -------
    T | D
        Item.

    Notes
    -----
    * If ``aiterable`` is empty and ``default`` is unset, then :obj:`IndexError` is raised.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await afirst(aiterable)
    0
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    item = await anext(aiterator, default)

    if item is ...:
        detail = "afirst(): empty iterable"
        raise IndexError(detail) from None

    return item



from collections.abc import AsyncIterable, AsyncIterator
from typing import TypeVar

from aioplus.internal.acount import acount
from aioplus.internal.azip import azip


T = TypeVar("T")


def aenumerate(aiterable: AsyncIterable[T], /, start: int = 0) -> AsyncIterator[tuple[int, T]]:
    """Enumerate ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    start : int, default 0
        Start.

    Returns
    -------
    AsyncIterator[tuple[int, T]]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(4, 23)
    >>> [(index, num) async for index, num in aenumerate(aiterable)]
    [(0, 4), (1, 5), (2, 6), (3, 7), ..., (17, 21), (18, 22)]

    See Also
    --------
    :func:`enumerate`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if not isinstance(start, int):
        detail = "'start' must be 'int'"
        raise TypeError(detail)

    return azip(acount(start), aiterable)



from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


def areversed(aiterable: AsyncIterable[T], /) -> AsyncIterator[T]:
    """Return reversed ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [num async for num in areversed(aiterable)]
    [22, 21, 20, 19, 18, ..., 4, 3, 2, 1, 0]

    See Also
    --------
    :func:`reversed`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    return AreversedIterator(aiterator)


@dataclass(repr=False)
class AreversedIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    aiterator: AsyncIterator[T]

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._started_flg: bool = False
        self._stack: list[T] = []

    async def __aioplus__(self) -> T:
        """Return the next item."""
        if not self._started_flg:
            self._started_flg = True
            self._stack = [item async for item in self.aiterator]

        if not self._stack:
            raise StopAsyncIteration

        return self._stack.pop()



from collections.abc import AsyncIterable
from typing import Any, Literal, Protocol, TypeVar, overload

from aioplus.internal.utils.typing import SupportsAdd, SupportsRAdd


AddableT1 = TypeVar("AddableT1", bound=SupportsAdd[Any, Any])
AddableT2 = TypeVar("AddableT2", bound=SupportsAdd[Any, Any])


class SupportsSumWithNoDefaultGiven(SupportsAdd[Any, Any], SupportsRAdd[int, Any], Protocol):
    """An ABC with two abstract methods `__add__` and `__radd__`."""


SupportsSumNoDefaultT = TypeVar("SupportsSumNoDefaultT", bound=SupportsSumWithNoDefaultGiven)


@overload
async def asum(aiterable: AsyncIterable[bool | int], /, *, start: int = 0) -> int: ...


@overload
async def asum(
    aiterable: AsyncIterable[SupportsSumNoDefaultT],
    /,
) -> SupportsSumNoDefaultT | Literal[0]: ...


@overload
async def asum(
    aiterable: AsyncIterable[AddableT1],
    /,
    *,
    start: AddableT2,
) -> AddableT1 | AddableT2: ...


async def asum(aiterable: AsyncIterable[Any], /, *, start: Any = 0) -> Any:
    """Sum ``start`` and items of ``aiterable`` from left to right.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    start : T
        Initializer.

    Returns
    -------
    T
        Sum.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await asum(aiterable)
    253

    See Also
    --------
    :func:`sum`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    total = start
    async for item in aiterable:
        total += item

    return total



from collections.abc import AsyncIterable
from typing import Any, TypeVar, overload

from aioplus.internal.aenumerate import aenumerate


T = TypeVar("T")
D = TypeVar("D")


@overload
async def anth(aiterable: AsyncIterable[T], /, *, n: int) -> T: ...


@overload
async def anth(aiterable: AsyncIterable[T], /, *, n: int, default: D) -> T | D: ...


async def anth(aiterable: AsyncIterable[Any], /, *, n: int, default: Any = ...) -> Any:
    """Return the ``n``-th item of ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    n : int
        Index.

    default : D, unset
        Default.

    Returns
    -------
    T | D
        Item.

    Notes
    -----
    * If ``aiterable[n]`` does not exist and ``default`` is unset, then :obj:`IndexError` is raised.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await anth(aiterable, n=4)
    4
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if not isinstance(n, int):
        detail = "'n' must be 'int'"
        raise TypeError(detail)

    if n < 0:
        detail = "'n' must be non-negative"
        raise ValueError(detail)

    async for index, item in aenumerate(aiterable):
        if index == n:
            return item

    if default is ...:
        detail = "'aiterable[n]' does not exist"
        raise IndexError(detail)

    return default



from collections.abc import AsyncIterator, Iterable, Iterator
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import partial
from typing import TypeVar

from aioplus.internal.awaitify import awaitify
from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


def anextify(
    iterable: Iterable[T],
    /,
    *,
    executor: ThreadPoolExecutor | None = None,
) -> AsyncIterator[T]:
    """Make ``iterable`` asynchronous.

    Parameters
    ----------
    iterable : Iterable[T]
        Iterable.

    executor : ThreadPoolExecutor, optional
        Executor.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Notes
    -----
    * If ``executor`` is :obj:`None`, then the default one is used (usually, a thread pool).

    Examples
    --------
    >>> iterable = [0, 1, 2, 3, 4, 5]
    >>> aiterable = anextify(iterable)
    >>> [num async for num in aiterable]
    [0, 1, 2, 3, 4, 5]

    See Also
    --------
    :meth:`asyncio.loop.run_in_executor`
    """
    if not isinstance(iterable, Iterable):
        detail = "'iterable' must be 'Iterable'"
        raise TypeError(detail)

    if executor is not None and not isinstance(executor, ThreadPoolExecutor):
        detail = "'executor' must be 'ThreadPoolExecutor' or 'None'"
        raise TypeError(detail)

    iterator = iter(iterable)
    return AnextifyIterator(iterator, executor)


@dataclass(repr=False)
class AnextifyIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    iterator: Iterator[T]
    executor: ThreadPoolExecutor | None

    async def __aioplus__(self) -> T:
        """Return the next item."""
        func = partial(next, self.iterator, ...)
        afunc = awaitify(func, executor=self.executor)

        item = await afunc()
        if item is ...:
            raise StopAsyncIteration

        return item



from aioplus.internal.aall import aall
from aioplus.internal.aany import aany
from aioplus.internal.abatched import abatched
from aioplus.internal.achain import achain
from aioplus.internal.acount import acount
from aioplus.internal.acycle import acycle
from aioplus.internal.aempty import aempty
from aioplus.internal.aenumerate import aenumerate
from aioplus.internal.afirst import afirst
from aioplus.internal.ahead import ahead
from aioplus.internal.aislice import aislice
from aioplus.internal.alast import alast
from aioplus.internal.alen import alen
from aioplus.internal.amax import amax
from aioplus.internal.amin import amin
from aioplus.internal.aminmax import aminmax
from aioplus.internal.anextify import anextify
from aioplus.internal.anth import anth
from aioplus.internal.apairwise import apairwise
from aioplus.internal.apostpend import apostpend
from aioplus.internal.aprepend import aprepend
from aioplus.internal.arange import arange
from aioplus.internal.arepeat import arepeat
from aioplus.internal.areversed import areversed
from aioplus.internal.asum import asum
from aioplus.internal.atabulate import atabulate
from aioplus.internal.atail import atail
from aioplus.internal.atriplewise import atriplewise
from aioplus.internal.awaitify import awaitify
from aioplus.internal.awindowed import awindowed
from aioplus.internal.azip import azip
from aioplus.internal.caller_thread_executor import CallerThreadExecutor


__author__ = "Sergei Y. Bogdanov <syubogdanov@outlook.com>"
__version__ = "0.7.0"

__all__: list[str] = [
    "CallerThreadExecutor",
    "aall",
    "aany",
    "abatched",
    "achain",
    "acount",
    "acycle",
    "aempty",
    "aenumerate",
    "afirst",
    "ahead",
    "aislice",
    "alast",
    "alen",
    "amax",
    "amin",
    "aminmax",
    "anextify",
    "anth",
    "apairwise",
    "apostpend",
    "aprepend",
    "arange",
    "arepeat",
    "areversed",
    "asum",
    "atabulate",
    "atail",
    "atriplewise",
    "awaitify",
    "awindowed",
    "azip",
]


CallerThreadExecutor.__module__ = "aioplus"
aall.__module__ = "aioplus"
aany.__module__ = "aioplus"
abatched.__module__ = "aioplus"
achain.__module__ = "aioplus"
acount.__module__ = "aioplus"
acycle.__module__ = "aioplus"
aempty.__module__ = "aioplus"
aenumerate.__module__ = "aioplus"
afirst.__module__ = "aioplus"
ahead.__module__ = "aioplus"
aislice.__module__ = "aioplus"
alast.__module__ = "aioplus"
alen.__module__ = "aioplus"
amax.__module__ = "aioplus"
amin.__module__ = "aioplus"
aminmax.__module__ = "aioplus"
anextify.__module__ = "aioplus"
anth.__module__ = "aioplus"
apairwise.__module__ = "aioplus"
apostpend.__module__ = "aioplus"
aprepend.__module__ = "aioplus"
arange.__module__ = "aioplus"
arepeat.__module__ = "aioplus"
areversed.__module__ = "aioplus"
asum.__module__ = "aioplus"
atabulate.__module__ = "aioplus"
atail.__module__ = "aioplus"
atriplewise.__module__ = "aioplus"
awaitify.__module__ = "aioplus"
awindowed.__module__ = "aioplus"
azip.__module__ = "aioplus"



from asyncio import TaskGroup
from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import Any, TypeVar, overload

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")


@overload
def azip(aiterable: AsyncIterable[T], /, *, strict: bool) -> AsyncIterator[tuple[T]]: ...


@overload
def azip(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    /,
    *,
    strict: bool = False,
) -> AsyncIterator[tuple[T1, T2]]: ...


@overload
def azip(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    /,
    *,
    strict: bool = False,
) -> AsyncIterator[tuple[T1, T2, T3]]: ...


@overload
def azip(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    aiterable4: AsyncIterable[T4],
    /,
    *,
    strict: bool = False,
) -> AsyncIterator[tuple[T1, T2, T3, T4]]: ...


@overload
def azip(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    aiterable4: AsyncIterable[T4],
    aiterable5: AsyncIterable[T5],
    /,
    *,
    strict: bool = False,
) -> AsyncIterator[tuple[T1, T2, T3, T4, T5]]: ...


@overload
def azip(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    aiterable4: AsyncIterable[T4],
    aiterable5: AsyncIterable[T5],
    aiterable6: AsyncIterable[T6],
    /,
    *,
    strict: bool = False,
) -> AsyncIterator[tuple[T1, T2, T3, T4, T5, T6]]: ...


@overload
def azip(*aiterables: AsyncIterable[T], strict: bool = False) -> AsyncIterator[tuple[T, ...]]: ...


def azip(*aiterables: AsyncIterable[Any], strict: bool = False) -> AsyncIterator[tuple[Any, ...]]:
    """Iterate over several iterables in parallel, producing tuples with an item from each one.

    Parameters
    ----------
    *aiterables : AsyncIterable[T]
        Iterables.

    strict : bool, default False
        Strictness.

    Returns
    -------
    AsyncIterator[tuple[T, ...]]
        Iterator.

    Notes
    -----
    * If ``strict`` is :obj:`True` and iterator lengths differ, then raises :exc:`ValueError`.

    Examples
    --------
    >>> xs = arange(42)
    >>> ys = arange(4, 23)
    >>> [(x, y) async for x, y in azip(xs, ys)]
    [(0, 4), (1, 5), (2, 6), ..., (18, 22)]
    """
    if not aiterables:
        detail = "'*aiterables' must be non-empty"
        raise ValueError(detail)

    for aiterable in aiterables:
        if not isinstance(aiterable, AsyncIterable):
            detail = "'*aiterables' must be 'AsyncIterable'"
            raise TypeError(detail)

    if not isinstance(strict, bool):
        detail = "'strict' must be 'bool'"
        raise TypeError(detail)

    aiterators = [aiter(aiterable) for aiterable in aiterables]
    return AzipIterator(aiterators, strict)


@dataclass(repr=False)
class AzipIterator(AioplusIterator[tuple[T, ...]]):
    """An asynchronous iterator."""

    aiterators: list[AsyncIterator[T]]
    strict: bool

    async def __aioplus__(self) -> tuple[T, ...]:
        """Return the next value."""
        async with TaskGroup() as task_group:
            coroutines = [anext(aiterator, ...) for aiterator in self.aiterators]
            tasks = [task_group.create_task(coroutine) for coroutine in coroutines]

        maybe_results = [task.result() for task in tasks]
        results = [result for result in maybe_results if result is not ...]

        if not results:
            raise StopAsyncIteration

        if self.strict and len(results) < len(self.aiterators):
            detail = "azip(): length mismatch"
            raise ValueError(detail)

        if len(results) < len(maybe_results):
            raise StopAsyncIteration

        return tuple(results)



from asyncio import iscoroutinefunction
from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


R = TypeVar("R")


def atabulate(afunc: Callable[[int], Awaitable[R]], /, *, start: int = 0) -> AsyncIterator[R]:
    """Return ``await afunc(0)``, ``await afunc(1)``, ``await afunc(2)``, etc.

    Parameters
    ----------
    afunc : Callable[[int], Awaitable[R]]
        Callable.

    Returns
    -------
    AsyncIterator[R]
        Iterator.

    Examples
    --------
    >>> afunc = awaitify(lambda x: x * x)
    >>> [num async for num in atabulate(afunc)]
    [0, 1, 4, 9, 16, 25, 36, 49, ...]
    """
    if not callable(afunc):
        detail = "'func' must be 'Callable'"
        raise TypeError(detail)

    if not iscoroutinefunction(afunc):
        detail = "'func' must be a coroutine function"
        raise TypeError(detail)

    if not isinstance(start, int):
        detail = "'start' must be 'int'"
        raise TypeError(detail)

    return AtabulateIterator(afunc, start)


@dataclass(repr=False)
class AtabulateIterator(AioplusIterator[R]):
    """An asynchronous iterator."""

    afunc: Callable[[int], Awaitable[R]]
    next: int

    async def __aioplus__(self) -> R:
        """Return the next item."""
        item = await self.afunc(self.next)
        self.next += 1
        return item



from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from types import EllipsisType
from typing import TypeVar, overload

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


@overload
def aislice(aiterable: AsyncIterable[T], stop: int, /) -> AsyncIterator[T]: ...


@overload
def aislice(aiterable: AsyncIterable[T], start: int, stop: int, /) -> AsyncIterator[T]: ...


@overload
def aislice(
    aiterable: AsyncIterable[T],
    start: int,
    stop: int,
    step: int,
    /,
) -> AsyncIterator[T]: ...


def aislice(
    aiterable: AsyncIterable[T],
    start: int,
    stop: int | EllipsisType = ...,
    step: int | EllipsisType = ...,
    /,
) -> AsyncIterator[T]:
    """Return selected items from ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    start : int
        Start.

    stop : int, unset
        Stop.

    step : int, unset
        Step.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(2003)
    >>> [num async for num in aislice(aiterable, 4, 23)]
    [4, 5, 6, 7, 8, ..., 20, 21, 22]

    See Also
    --------
    :func:`itertools.islice`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if not isinstance(start, int):
        detail = "'start' must be 'int'"
        raise TypeError(detail)

    if (stop is not ...) and not isinstance(stop, int):
        detail = "'stop' must be 'int'"
        raise TypeError(detail)

    if (step is not ...) and not isinstance(step, int):
        detail = "'step' must be 'int'"
        raise TypeError(detail)

    if (stop is ...) and (step is not ...):
        detail = "'step' must be 'int'"
        raise TypeError(detail)

    if stop is ...:
        stop = start
        start = 0
        step = 1

    if step is ...:
        step = 1

    if start < 0:
        detail = "'start' must be non-negative"
        raise ValueError(detail)

    if stop < 0:
        detail = "'stop' must be non-negative"
        raise ValueError(detail)

    if step <= 0:
        detail = "'step' must be positive"
        raise ValueError(detail)

    aiterator = aiter(aiterable)
    return AisliceIterator(aiterator, start, stop, step)


@dataclass(repr=False)
class AisliceIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    aiterator: AsyncIterator[T]
    start: int
    stop: int
    step: int

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._next = 0
        self._yield_at = self.start

    async def __aioplus__(self) -> T:
        """Return the next item."""
        if self._yield_at >= self.stop:
            raise StopAsyncIteration

        for _ in range(self._yield_at - self._next):
            await anext(self.aiterator)
            self._next += 1

        item = await anext(self.aiterator)

        self._next += 1
        self._yield_at += self.step

        return item



from collections.abc import AsyncIterable, AsyncIterator
from typing import TypeVar

from aioplus.internal.awindowed import awindowed


T = TypeVar("T")


def atriplewise(aiterable: AsyncIterable[T], /) -> AsyncIterator[tuple[T, T, T]]:
    """Return a sliding window of width ``n=3`` over ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    Returns
    -------
    AsyncIterator[tuple[T, T, T]]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [triplet async for triplet in atriplewise(aiterable)]
    [(0, 1, 2), (1, 2, 3), ..., (19, 20, 21), (20, 21, 22)]
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    return awindowed(aiterable, n=3)



from collections.abc import AsyncIterable, Callable
from typing import Any, TypeAlias, TypeVar, overload

from aioplus.internal.utils.typing import SupportsDunderGT, SupportsDunderLT


T = TypeVar("T")
D = TypeVar("D")

SupportsRichComparison: TypeAlias = SupportsDunderLT[Any] | SupportsDunderGT[Any]
SupportsRichComparisonT = TypeVar("SupportsRichComparisonT", bound=SupportsRichComparison)


@overload
async def amin(
    iterable: AsyncIterable[SupportsRichComparisonT],
    /,
    *,
    key: None = None,
) -> SupportsRichComparisonT: ...


@overload
async def amin(
    iterable: AsyncIterable[T],
    /,
    *,
    key: Callable[[T], SupportsRichComparison],
) -> T: ...


@overload
async def amin(
    iterable: AsyncIterable[SupportsRichComparisonT],
    /,
    *,
    key: None = None,
    default: T,
) -> SupportsRichComparisonT | T: ...


@overload
async def amin(
    iterable: AsyncIterable[T],
    /,
    *,
    key: Callable[[T], SupportsRichComparison],
    default: D,
) -> T | D: ...


async def amin(
    aiterable: AsyncIterable[Any],
    /,
    *,
    key: Callable[[Any], Any] | None = None,
    default: Any = ...,
) -> Any:
    """Return the smallest item in ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    key : Callable[[T], SupportsRichComparison], optional
        Comparator.

    default : D, unset
        Default.

    Returns
    -------
    T | D
        Item.

    Notes
    -----
    * If ``aiterable`` is empty and ``default`` is unset, then :obj:`ValueError` is raised.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await amin(aiterable)
    0

    See Also
    --------
    :func:`min`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if key is not None and not callable(key):
        detail = "'key' must be 'Callable' or 'None'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    smallest = await anext(aiterator, ...)

    if smallest is not ...:
        async for item in aiterator:
            smallest = min(smallest, item, key=key)
        return smallest

    if default is ...:
        detail = "amin(): empty iterable"
        raise ValueError(detail) from None

    return default



from collections import deque
from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


def atail(aiterable: AsyncIterable[T], /, *, n: int) -> AsyncIterator[T]:
    """Return the last ``n`` items of the ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    n : int
        Count.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [num async for num in atail(aiterable, n=4)]
    [19, 20, 21, 22]
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if not isinstance(n, int):
        detail = "'n' must be 'int'"
        raise TypeError(detail)

    if n < 0:
        detail = "'n' must be non-negative"
        raise ValueError(detail)

    aiterator = aiter(aiterable)
    return AtailIterator(aiterator, n)


@dataclass(repr=False)
class AtailIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    aiterator: AsyncIterable[T]
    n: int

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._started_flg: bool = False
        self._deque: deque[T] = deque(maxlen=self.n)

    async def __aioplus__(self) -> T:
        """Return the next item."""
        if not self._started_flg:
            self._started_flg = True
            async for item in self.aiterator:
                self._deque.append(item)

        if not self._deque:
            raise StopAsyncIteration

        return self._deque.popleft()



from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar, overload

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")

T1 = TypeVar("T1")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
T4 = TypeVar("T4")
T5 = TypeVar("T5")
T6 = TypeVar("T6")


@overload
def achain(aiterable: AsyncIterable[T], /) -> AsyncIterator[T]: ...


@overload
def achain(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    /,
) -> AsyncIterator[T1 | T2]: ...


@overload
def achain(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    /,
) -> AsyncIterator[T1 | T2 | T3]: ...


@overload
def achain(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    aiterable4: AsyncIterable[T4],
    /,
) -> AsyncIterator[T1 | T2 | T3 | T4]: ...


@overload
def achain(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    aiterable4: AsyncIterable[T4],
    aiterable5: AsyncIterable[T5],
    /,
) -> AsyncIterator[T1 | T2 | T3 | T4 | T5]: ...


@overload
def achain(
    aiterable1: AsyncIterable[T1],
    aiterable2: AsyncIterable[T2],
    aiterable3: AsyncIterable[T3],
    aiterable4: AsyncIterable[T4],
    aiterable5: AsyncIterable[T5],
    aiterable6: AsyncIterable[T6],
    /,
) -> AsyncIterator[T1 | T2 | T3 | T4 | T5 | T6]: ...


@overload
def achain(*aiterables: AsyncIterable[T]) -> AsyncIterator[T]: ...


def achain(*aiterables: AsyncIterable[T]) -> AsyncIterator[T]:
    """Iterate ``*aiterables`` sequentially.

    Parameters
    ----------
    *aiterables : AsyncIterable[T]
        Iterables.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Examples
    --------
    >>> nums1 = arange(0, 3)
    >>> nums2 = arange(3, 6)
    >>> [num async for num in achain(nums1, nums2)]
    [0, 1, 2, 3, 4, 5]
    """
    if not aiterables:
        detail = "'*aiterables' must be non-empty"
        raise ValueError(detail)

    for aiterable in aiterables:
        if not isinstance(aiterable, AsyncIterable):
            detail = "'*aiterables' must be 'AsyncIterable'"
            raise TypeError(detail)

    aiterators = [aiter(aiterable) for aiterable in aiterables]
    return AchainIterator(aiterators)


@dataclass(repr=False)
class AchainIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    aiterators: list[AsyncIterator[T]]

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._index = 0

    async def __aioplus__(self) -> T:
        """Return the next item."""
        while self._index < len(self.aiterators):
            aiterator = self.aiterators[self._index]
            try:
                item = await anext(aiterator)

            except StopAsyncIteration:
                self._index += 1
                continue

            return item

        raise StopAsyncIteration



from collections.abc import AsyncIterable
from typing import Any


async def alen(aiterable: AsyncIterable[Any], /) -> int:
    """Return length of ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[Any]
        Iterable.

    Returns
    -------
    :class:`int`
        Length.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await alen(aiterable)
    23

    See Also
    --------
    :func:`len`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    count = 0
    async for _ in aiterable:
        count += 1

    return count



from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.aislice import aislice
from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


def abatched(
    aiterable: AsyncIterable[T],
    /,
    *,
    n: int,
    strict: bool = False,
) -> AsyncIterator[tuple[T, ...]]:
    """Iterate ``aiterable`` by batches of length ``n``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    n : int
        Batch size.

    strict : bool, default False
        Strictness.

    Returns
    -------
    AsyncIterator[tuple[T, ...]]
        Iterator.

    Notes
    -----
    * If ``strict`` is :obj:`True` and the total number of objects is not divisible by ``n``, then
      raises :exc:`ValueError`. If :obj:`False`, the last batch may be shorter than ``n``.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [batch async for batch in abatched(aiterable, n=3)]
    [(0, 1, 2), (3, 4, 5), ..., (18, 19, 20), (21, 22)]

    See Also
    --------
    :func:`itertools.batched`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if not isinstance(n, int):
        detail = "'n' must be 'int'"
        raise TypeError(detail)

    if n <= 0:
        detail = "'n' must be positive"
        raise ValueError(detail)

    if not isinstance(strict, bool):
        detail = "'strict' must be 'bool'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    return AbatchedIterator(aiterator, n, strict)


@dataclass
class AbatchedIterator(AioplusIterator[tuple[T, ...]]):
    """An asynchronous iterator."""

    aiterator: AsyncIterator[T]
    n: int
    strict: bool

    async def __aioplus__(self) -> tuple[T, ...]:
        """Return the next item."""
        batch = [item async for item in aislice(self.aiterator, self.n)]

        if not batch:
            raise StopAsyncIteration

        if self.strict and len(batch) < self.n:
            detail = "abatched(): incomplete batch"
            raise ValueError(detail)

        return tuple(batch)



from collections.abc import AsyncIterable, AsyncIterator
from typing import TypeVar

from aioplus.internal.awindowed import awindowed


T = TypeVar("T")


def apairwise(aiterable: AsyncIterable[T], /) -> AsyncIterator[tuple[T, T]]:
    """Return a sliding window of width ``n=2`` over ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    Returns
    -------
    AsyncIterator[tuple[T, T]]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [pair async for pair in apairwise(aiterable)]
    [(0, 1), (1, 2), (2, 3), ..., (20, 21), (21, 22)]

    See Also
    --------
    :func:`itertools.pairwise`
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    return awindowed(aiterable, n=2)



from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")
V = TypeVar("V")


def apostpend(aiterable: AsyncIterable[T], value: V, /) -> AsyncIterator[T | V]:
    """Yield from ``aiterable``, then ``value``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    value : V
        Value.

    Returns
    -------
    AsyncIterator[T | V]
        Iterator.

    Examples
    --------
    >>> [num async for num in apostpend(arange(4), 4)]
    [0, 1, 2, 3, 4]
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    return ApostpendIterator(aiterator, value)


@dataclass(repr=False)
class ApostpendIterator(AioplusIterator[T | V]):
    """An asynchronous iterator."""

    aiterator: AsyncIterator[T]
    value: V

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._finished_flg: bool = False

    async def __aioplus__(self) -> T | V:
        """Return the next item."""
        if self._finished_flg:
            raise StopAsyncIteration

        try:
            item = await anext(self.aiterator)

        except StopAsyncIteration:
            self._finished_flg = True
            return self.value

        return item



from asyncio import get_running_loop
from collections.abc import Awaitable, Callable
from concurrent.futures import ThreadPoolExecutor
from contextvars import copy_context
from functools import partial, wraps
from typing import ParamSpec, TypeVar


P = ParamSpec("P")
R = TypeVar("R")


def awaitify(
    func: Callable[P, R],
    /,
    *,
    executor: ThreadPoolExecutor | None = None,
) -> Callable[P, Awaitable[R]]:
    """Make ``func`` asynchronous.

    Parameters
    ----------
    func : Callable[P, R]
        Callable.

    executor : ThreadPoolExecutor, optional
        Executor.

    Returns
    -------
    Callable[P, Awaitable[R]]
        Callable.

    Notes
    -----
    * If ``executor`` is :obj:`None`, then the default one is used (usually, a thread pool).

    Examples
    --------
    >>> aprint = awaitify(print)
    >>> await aprint("4 -> 23")
    4 -> 23

    See Also
    --------
    :meth:`asyncio.loop.run_in_executor`
    """
    if not callable(func):
        detail = "'func' must be 'Callable'"
        raise TypeError(detail)

    if executor is not None and not isinstance(executor, ThreadPoolExecutor):
        detail = "'executor' must be 'ThreadPoolExecutor' or 'None'"
        raise TypeError(detail)

    @wraps(func)
    async def afunc(*args: P.args, **kwargs: P.kwargs) -> R:
        """Run the function asynchronously."""
        loop = get_running_loop()
        context = copy_context()
        execute = partial(context.run, func, *args, **kwargs)
        return await loop.run_in_executor(executor, execute)

    return afunc



from typing import Protocol, Self, TypeVar, runtime_checkable


T_co = TypeVar("T_co", covariant=True)
T_contra = TypeVar("T_contra", contravariant=True)


@runtime_checkable
class SupportsAdd(Protocol[T_contra, T_co]):
    """An ABC with one abstract method `__add__`."""

    def __add__(self, x: T_contra, /) -> T_co:
        """Perform addition operation."""


@runtime_checkable
class SupportsBool(Protocol):
    """An ABC with one abstract method `__bool__`."""

    def __bool__(self) -> bool:
        """Perform boolean conversion."""


@runtime_checkable
class SupportsRAdd(Protocol[T_contra, T_co]):
    """An ABC with one abstract method `__radd__`."""

    def __radd__(self, x: T_contra, /) -> T_co:
        """Perform reverse addition operation."""


@runtime_checkable
class SupportsDunderLT(Protocol[T_contra]):
    """An ABC with one abstract method `__lt__`."""

    def __lt__(self, other: T_contra, /) -> bool:
        """Perform less-than comparison."""


@runtime_checkable
class SupportsDunderGT(Protocol[T_contra]):
    """An ABC with one abstract method `__gt__`."""

    def __gt__(self, other: T_contra, /) -> bool:
        """Perform greater-than comparison."""


@runtime_checkable
class AcloseableIterator(Protocol[T_co]):
    """An ABC that implements three abstract methods: `__aiter__`, `__anext__` and `aclose`."""

    def __aiter__(self) -> Self:
        """Return self."""

    async def __anext__(self) -> T_co:
        """Return the next item."""

    async def aclose(self) -> object:
        """Close the object."""



from collections.abc import AsyncIterable
from typing import Any


async def aempty(aiterable: AsyncIterable[Any], /) -> bool:
    """Return :obj:`True` if ``aiterable`` is empty, otherwise :obj:`False`.

    Parameters
    ----------
    aiterable : AsyncIterable[Any]
        Iterable.

    Returns
    -------
    :class:`bool`
        Emptiness.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> await aempty(aiterable)
    False
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    try:
        await anext(aiterator)

    except StopAsyncIteration:
        return True

    return False



from collections.abc import AsyncIterator
from dataclasses import dataclass

from aioplus.internal.utils.abc import AioplusIterator


def acount(start: int = 0, step: int = 1) -> AsyncIterator[int]:
    """Return evenly spaced integers.

    Parameters
    ----------
    start : int, default 0
        Initializer.

    step : int, default 1
        Step.

    Returns
    -------
    AsyncIterator[int]
        Iterator.

    Examples
    --------
    >>> [num async for num in acount(start=23, step=4)]
    [23, 27, 31, 35, 39, 43, 47, ...]

    See Also
    --------
    :func:`itertools.count`
    """
    if not isinstance(start, int):
        detail = "'start' must be 'int'"
        raise TypeError(detail)

    if not isinstance(step, int):
        detail = "'step' must be 'int'"
        raise TypeError(detail)

    return AcountIterator(start, step)


@dataclass(repr=False)
class AcountIterator(AioplusIterator[int]):
    """An asynchronous iterator."""

    next: int
    step: int

    async def __aioplus__(self) -> int:
        """Return the next item."""
        item = self.next
        self.next += self.step
        return item



from collections.abc import AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


def arepeat(obj: T, /, *, times: int | None = None) -> AsyncIterator[T]:
    """Return the same object repeatedly.

    Parameters
    ----------
    obj : T
        Object.

    times : int, optional
        Count.

    Returns
    -------
    AsyncIterator[T]
        Iterator.

    Notes
    -----
    * If ``times`` is :obj:`None`, then the iterable will be infinite.

    Examples
    --------
    >>> [num async for num in arepeat(23, times=4)]
    [23, 23, 23, 23]

    See Also
    --------
    :func:`itertools.repeat`
    """
    if times is not None and not isinstance(times, int):
        detail = "'times' must be 'int' or 'None'"
        raise TypeError(detail)

    if times is not None and times < 0:
        detail = "'times' must be non-negative"
        raise ValueError(detail)

    return ArepeatIterator(obj, times)


@dataclass(repr=False)
class ArepeatIterator(AioplusIterator[T]):
    """An asynchronous iterator."""

    obj: T
    times: int | None

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._count: int = 0

    async def __aioplus__(self) -> T:
        """Return the next item."""
        if self.times is None:
            return self.obj

        if self._count >= self.times:
            raise StopAsyncIteration

        self._count += 1
        return self.obj



from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import TypeVar

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")
V = TypeVar("V")


def aprepend(value: V, aiterable: AsyncIterable[T], /) -> AsyncIterator[V | T]:
    """Yield ``value``, then from ``aiterable``.

    Parameters
    ----------
    value : V
        Value.

    aiterable : AsyncIterable[T]
        Iterable.

    Returns
    -------
    AsyncIterator[V | T]
        Iterator.

    Examples
    --------
    >>> [num async for num in aprepend(0, arange(1, 5))]
    [0, 1, 2, 3, 4]
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    aiterator = aiter(aiterable)
    return AprependIterator(value, aiterator)


@dataclass(repr=False)
class AprependIterator(AioplusIterator[V | T]):
    """An asynchronous iterator."""

    value: V
    aiterator: AsyncIterator[T]

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._started_flg: bool = False

    async def __aioplus__(self) -> V | T:
        """Return the next item."""
        if not self._started_flg:
            self._started_flg = True
            return self.value

        return await anext(self.aiterator)



from collections import deque
from collections.abc import AsyncIterable, AsyncIterator
from dataclasses import dataclass
from typing import Literal, TypeVar, overload

from aioplus.internal.utils.abc import AioplusIterator


T = TypeVar("T")


@overload
def awindowed(aiterable: AsyncIterable[T], /, *, n: Literal[2]) -> AsyncIterator[tuple[T, T]]: ...


@overload
def awindowed(
    aiterable: AsyncIterable[T],
    /,
    *,
    n: Literal[3],
) -> AsyncIterator[tuple[T, T, T]]: ...


@overload
def awindowed(aiterable: AsyncIterable[T], /, *, n: int) -> AsyncIterator[tuple[T, ...]]: ...


def awindowed(aiterable: AsyncIterable[T], /, *, n: int) -> AsyncIterator[tuple[T, ...]]:
    """Return a sliding window of width ``n`` over ``aiterable``.

    Parameters
    ----------
    aiterable : AsyncIterable[T]
        Iterable.

    n : int
        Width.

    Returns
    -------
    AsyncIterator[tuple[T, ...]]
        Iterator.

    Examples
    --------
    >>> aiterable = arange(23)
    >>> [window async for window in awindowed(aiterable, n=3)]
    [(0, 1, 2), (1, 2, 3), ..., (19, 20, 21), (20, 21, 22)]
    """
    if not isinstance(aiterable, AsyncIterable):
        detail = "'aiterable' must be 'AsyncIterable'"
        raise TypeError(detail)

    if not isinstance(n, int):
        detail = "'n' must be 'int'"
        raise TypeError(detail)

    if n <= 0:
        detail = "'n' must be positive"
        raise ValueError(detail)

    aiterator = aiter(aiterable)
    return AwindowedIterator(aiterator, n)


@dataclass(repr=False)
class AwindowedIterator(AioplusIterator[tuple[T, ...]]):
    """An asynchronous iterator."""

    aiterator: AsyncIterator[T]
    n: int

    def __post_init__(self) -> None:
        """Initialize the object."""
        self._window: deque[T] = deque(maxlen=self.n)

    async def __aioplus__(self) -> tuple[T, ...]:
        """Return the next item."""
        while len(self._window) < self.n - 1:
            item = await anext(self.aiterator)
            self._window.append(item)

        item = await anext(self.aiterator)

        self._window.append(item)
        return tuple(self._window)
